// Copyright (C) 2018-2019  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// XXX doc

#ifndef _NXD_WCFS_H_
#define _NXD_WCFS_H_

#include <golang/libgolang.h>
#include <golang/cxx.h>
#include <golang/sync.h>
using namespace golang;
using cxx::dict;
using cxx::set;

#include <string>
using std::string;

#include <tuple>
using std::tuple;

#include <utility>
using std::pair;

#include "wcfs_misc.h"

// wcfs::
namespace wcfs {

struct _File;
struct _Mapping;
struct PinReq;
typedef refptr<class _Conn> Conn;
typedef refptr<class _WatchLink> WatchLink;


// WCFS represents filesystem-level connection to wcfs server.
// XXX doc
struct WCFS {
    string  mountpoint;

    pair<Conn, error>       connect(zodb::Tid at);
    string                  _path(const string &obj);
    tuple<os::File, error>  _open(const string &path, int flags=O_RDONLY);
    pair<WatchLink, error>  _openwatch();
};

// Conn represents logical connection that provides view of data on wcfs
// filesystem as of particular database state.
//
// It uses /head/bigfile/* and notifications received from /head/watch to
// maintain isolated database view while at the same time sharing most of data
// cache in OS pagecache of /head/bigfile/*.
//
// Use WCFS.connect(at) to create Conn.
// Use .mmap to create new Mappings.
// Use .resync to resync Conn onto different database view.
//
// Conn logically mirrors ZODB.Connection .
typedef refptr<struct _Conn> Conn;
struct _Conn : object {
    WCFS        *_wc;
    zodb::Tid   at;
    WatchLink   _wlink; // watch/receive pins for created mappings

    sync::Mutex              _filemu;
    dict<zodb::Oid, _File*>  _filetab; // {} foid -> _file

    sync::WorkGroup _pinWG;
    func<void()>    _pinCancel;

    // don't new - create via WCFS.connect
private:
    _Conn();
    ~_Conn();
    friend pair<Conn, error> WCFS::connect(zodb::Tid at);
public:
    void decref();

public:
    error close();
    // XXX move mmap -> _FileH ?
    pair<_Mapping*, error> mmap(zodb::Oid foid, int64_t blk_start, int64_t blk_len);
    error resync(zodb::Tid at);

private:
    error _pinner(context::Context ctx);
    void _pin1(PinReq *req);
};


}   // wcfs::

#endif
